import argparse

argparser = argparse.ArgumentParser()
argparser.add_argument("-l", type=int, help="length", default=1)
argparser.add_argument("-x", type=int, help="height", default=2)
argparser.add_argument("-w", type=int, help="width", default=3)
argparser.add_argument("-s", type=str, help="shape", default="box")
args = argparser.parse_args()

def main():

    green_code = "\033[92m" # Green
    reset_code = "\033[0m"  # Reset color
    red_code = "\033[91m"   # Red

    volume = args.l * args.x * args.w
    shape = args.s
    print(f"{red_code}Length: {args.l}, Height: {args.x}, Width: {args.w}{reset_code}")
    print(f"Volume of {shape} is {volume}")

    print(f"{green_code}TEST SUCCESFUL!{reset_code}")
    # prb test -l 10 -x 20 -w 30 -s box
if __name__ == "__main__":

    main()


#!/usr/bin/python3

from pickle import FALSE, TRUE
import numpy as np
from tabulate import tabulate as tab
import os
from tqdm import tqdm
import pandas as pd
import sys

types = {'DNA' : 'Reference', 'RNA' : 'RevCompl', '-RNA' : 'Reference'}

def reform_hush(nt_type:str='DNA',
                sub:bool=False,L:int=80,l:int=22,
                currentfolder:os.PathLike = './data')->None:

    suffix = types[nt_type] # see types dict above
    roilist = currentfolder+'/rois/all_regions.tsv'
    rd = pd.read_csv(roilist,sep="\t",header=0)
    ROIcount = len(rd)

    infolder = currentfolder+'/candidates/'
    out = currentfolder+'/HUSH_candidates/'

    try:
        os.mkdir(out)
    except FileExistsError:
        print("Saving to existing 'HUSH_candidates' directory.")

    if (sub==TRUE):
        # each oligo gets an off-target score based on the min number of mismatch in any of the sublengths, 
        # and the sum of mismatches count across all sublength oligos

        for k1 in tqdm(range(ROIcount), 'Processing all ROI'):
            filename = 'roi_'+str(k1+1)+'.GC35to85_'+suffix+'.fa'
            fasta = infolder+filename
            hush = fasta+'.'+str(l)+'mers.nh.L'+str(l)+'.mindist.uint8'
            hdist = np.fromfile(hush,'uint8')

            n = len(hdist)
            #print('HUSH list: '+str(n))
            hdist_grouped_min = hdist.reshape([n//(L-l+1), (L-l+1)]).min(axis=1)
            hdist_grouped_sum = hdist.reshape([n//(L-l+1), (L-l+1)]).sum(axis=1)
            #print('Grouped length: '+str(len(hdist_grouped_min)))
                    
            f = open(fasta,'r')
            full = f.read()
            splitseq = full.splitlines()

            for k2 in range(len(hdist_grouped_min)):
                #print('Current row: '+str(k2))
                splitseq[2*k2] = splitseq[2*k2] + "\n"
                # /!\ UGLY SOLUTION
                # encode the two mismatch parameters (min mismatch hit and sum mismatch count) in a format compatible with ifpd2 db make
                splitseq[2*k2+1] = splitseq[2*k2+1] + ", " + "111" + str(hdist_grouped_min[k2]) + "987" + str(hdist_grouped_sum[k2]) + "\n"
                    
            o = open(out+filename,'w')
            outseq = ''.join(splitseq)
            o.write(outseq)
            o.close()
            f.close()  

    else:
        # each oligo gets an off-target score based on the min number of mismatches as found by nHUSH

        for k1 in tqdm(range(ROIcount), 'Processing all ROI'):
            filename = 'roi_'+str(k1+1)+'.GC35to85_'+suffix+'.fa'
            fasta = infolder+filename
            hush = fasta+'.nh.L'+str(L)+'.mindist.uint8'
            hdist = np.fromfile(hush,'uint8')

            n = len(hdist)                    
            f = open(fasta,'r')
            full = f.read()
            splitseq = full.splitlines()

            for k2 in range(n):
                #print('Current row: '+str(k2))
                splitseq[2*k2] = splitseq[2*k2] + "\n"
                # /!\ UGLY SOLUTION
                # encode the two mismatch parameters (min mismatch hit and sum mismatch count) in a format compatible with ifpd2 db make
                splitseq[2*k2+1] = splitseq[2*k2+1] + ", " + "111" + str(hdist[k2]) + "98799"  + "\n"
                    
            o = open(out+filename,'w')
            outseq = ''.join(splitseq)
            o.write(outseq)
            o.close()
            f.close()
    return

if __name__ == "__main__":
    #syntax: ./reform_hush.py DNA/RNA/-RNA 80 (22)

    nt_type = sys.argv[1] # DNA/RNA/-RNA
    print(f'Number of arguments: '+str(len(sys.argv)))
    if(len(sys.argv) == 3):
        # no sublength was used
        sub = FALSE
    elif(len(sys.argv) == 4):
        sub = TRUE
    else:
         print(f'Incorrect number of arguments. Exiting...')
         exit(-1)

    L = int(sys.argv[2])     #full oligo length
    print(f'Length: '+str(L))
    if (sub==TRUE):
        l = int(sys.argv[3]) # sublength
        print(f'Sublength: '+str(l))
        reform_hush(nt_type=nt_type,sub=sub,L=L,l=l)
    else:
        print(f'No sublength used')
        reform_hush(nt_type=nt_type,sub=sub,L=L)
    



# Python code to download the chromosome 17 of the human genome from https://ftp.ensembl.org/pub/release-110/fasta/homo_sapiens/dna/Homo_sapiens.GRCh38.dna.chromosome.17.fa.gz

import requests
import os
import gzip

# Human (homo_sapiens) and Mouse (mus_musculus) genomes respectively
GRCh = "https://ftp.ensembl.org/pub/release-{release}/fasta/homo_sapiens/dna/Homo_sapiens.GRCh{build}.dna.chromosome.{chr}.fa.gz"
GRCm = "https://ftp.ensembl.org/pub/release-{release}/fasta/mus_musculus/dna/Mus_musculus.GRCm{build}.dna.chromosome.{chr}.fa.gz"

# available only mouse and human
available = ['human','homo_sapiens','mouse','mus_musculus']
human = ['human','homo_sapiens']
mouse = ['mouse','mus_musculus']

human_all_chr: list[int|str] = list(range(1,23))+['X','Y'] # list of human chromosomes
mouse_all_chr: list[int|str] = list(range(1,19))+['X','Y'] # list of mouse chromosomes

def download_chr(chr_folder:os.PathLike="data/ref",
                 chr:int|str=17,
                 release:int|str=109,
                 build:int=38,
                 species:str="human",
                 unzip:bool=True,
                 remove_gz:bool=False,
                 file_name:str|None=None,
                 url_chr:str|None=None)->None:
    
    """
    Downloads the specified chromosome of a genome from Ensembl FTP server. Requires internet connection...

    Args:
        chr_folder (os.PathLike, optional): 
            Path to the folder where the chromosome file will be saved. Defaults to "data/ref".
        chr (int|str, optional): 
            Chromosome number or identifier. Defaults to 17.
        release (int|str, optional): 
            Release version of the genome. Defaults to 109.
        build (int, optional):
            Build version of the genome. Defaults to 38.
        species (str, optional):
            Species of the genome (human or mouse). Defaults to "human".
        unzip (bool, optional): 
            Flag to indicate whether to unzip the downloaded file. Defaults to True.
        remove_gz (bool, optional): 
            Flag to indicate whether to remove the downloaded gzipped file. Defaults to False.

    Returns:
        None
    """

    # matching species with the link
    species = species.lower()
    if url_chr is None:
        if species in human:
            url_chr = GRCh
        elif species in mouse:
            url_chr = GRCm
        else:
            print(f"species should be one of {available}")
            print(f"You entered {species}")
            return
        print(f"downloading {species} chromosome {chr}...")
        url_chr = url_chr.replace('{release}',f'{release}'
                                    ).replace('{chr}',f'{chr}'
                                        ).replace('{build}',f'{build}') # placing release and chr numbers
    else:
        print(f"downloading from {url_chr}...")
        print(f"if reference genome, this may take some time: 5-15 mins depending on your internet speed")
        pass # url_chr = url_chr

    if file_name is None:
        file_name = url_chr.split('/')[-1] # last part of the link will be the file name
    else:
        pass # file_name = file_name
    
    # creates chr_folder if not exists
    if not os.path.exists(chr_folder):
        try: # just in case the parent directories do not exist
            os.mkdir(chr_folder)
        except:
            print(f'parent directory {os.path.dirname(chr_folder)} does not exist')
            print('exiting...')
            return
    
    chr_path = os.path.join(chr_folder,file_name)
    
    # read the link to a variable chr_fasta_gz
    if not os.path.exists(chr_path): # do not overwrite if it exists
        
        try:
            chr_fasta_gz = requests.get(url_chr)
            open(chr_path,"wb").write(chr_fasta_gz.content)# write chr_fasta_gz into the chromosome path
        except:
            print("Link does not work:",url_chr)
            print("Exiting...")
            return
    else:
        print(f'file {chr_path} already exists. Skipping...')


    if unzip: # unzip the .gz file if requested
        chr_fasta = chr_path[:-3]  # Remove the '.gz' extension
        if not os.path.exists(chr_fasta): # do not overwrite chromosome fasta file if exists
            print(f'unzipping...')
            # Unzip .gz file and save the unzipped content to a new file
            open(chr_fasta, 'wb').write(gzip.open(chr_path, 'rb').read())
            print(f'unzipped file saved to {chr_fasta}')
        else:
            print(f'file {chr_fasta} already exists. Skipping...')

    if remove_gz: # remove the .gz file if requested
        if os.path.exists(chr_path):
            print(f'Removing gzipped file {chr_path} ...')
            os.remove(chr_path)
            print('Done')
        else: # could already be removed
            print(f'file {chr_path} does not exist. Skipping...')


    return

def download_chr_list(list_chr:str|list = 'all',   
                      chr_folder:os.PathLike="data/ref",
                      release:int|str=109,
                      build:int=38,
                      species:str="human",
                      unzip:bool=True,
                      remove_gz:bool=False)->None:
    
    """
    Downloads the specified chromosome lists of a genome from Ensembl FTP server. 
        Requires internet connection...

    Args:
        list_chr (list[int|str], optional): 
            List of chromosome numbers or identifiers. Defaults to list(range(1,23))+['X','Y'].
        chr_folder (os.PathLike, optional): 
            Path to the folder where the chromosome file will be saved. Defaults to "data/ref".
        release (int|str, optional): 
            Release version of the genome. Defaults to 109.
        build (int, optional):
            Build version of the genome. Defaults to 38.
        species (str, optional):
            Species of the genome (human or mouse). Defaults to "human".
        unzip (bool, optional): 
            Flag to indicate whether to unzip the downloaded file. Defaults to True.
        remove_gz (bool, optional): 
            Flag to indicate whether to remove the downloaded gzipped file. Defaults to False.

    Returns:
        None
    """
    if list_chr=="all":
        if species in human :
            list_chr = human_all_chr
        elif species in mouse:
            list_chr = mouse_all_chr
        else:
            print(f"species should be one of {available}")
            print(f"You entered {species}")
            return
    elif type(list_chr) == list or type(list_chr) == tuple:
        pass # 
    else:
        print('list_chr should either be "all", a list or tuple')
        print(f'You entered {type(list_chr)} type which is not comptatible..')
        return

    for chr in list_chr:
        print(f'downloanding... {species=} {chr=} {release=} {build=} ')
        download_chr(chr_folder=chr_folder,
                     chr=chr,
                     release=release,
                     species=species,
                     unzip=unzip,
                     remove_gz=remove_gz,
                     )
    return

def download_ref_genome(build:int=38,
                        release:int=109,
                        species:str="human",
                        chr_folder:os.PathLike="data/ref",
                        file_name:str='genome.fa.gz',
                        unzip:bool=True,
                        remove_gz:bool=False)->None:
    
    if species in human:
        url_chr = GRCh.replace('{release}',f'{release}').replace('{build}',f'{build}')
        
    elif species in mouse:
        url_chr = GRCm.replace('{release}',f'{release}').replace('{build}',f'{build}')


    url_chr = url_chr.replace('chromosome.{chr}','primary_assembly')
    download_chr(chr_folder=chr_folder,
                 url_chr=url_chr,
                 unzip=unzip,
                 file_name=file_name,
                 remove_gz=remove_gz)
    return



if __name__ == "__main__":
    download_ref_genome() # works for the default values


#!/usr/bin/env python

from Bio.SeqIO.FastaIO import SimpleFastaParser
import numpy as np
import pandas as pd
import os
import glob
import sys
import click
import re
from tqdm import tqdm 
from tabulate import tabulate
# import statistics as stat
import shutil
from itertools import compress
import subprocess
import logging
from datetime import datetime
import multiprocessing as mp
from joblib import Parallel, delayed
import joblib
import contextlib

pd.options.mode.chained_assignment = None  # default='warn'. Suppress SettingWithCopyWarning

@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()

@click.command(
    name="cycling_query",
    help="Generate an optimized probe set for each ROI."
)
@click.option('-s', '--strand', type=click.STRING)
@click.option('-L', '--length', type=click.INT)
@click.option('-m', '--mismatch', type=click.INT)
@click.option('-c', '--cutoff', type=click.INT)
@click.option('-t', '--threads', type=click.INT)
@click.option('-g', '--gap', type=click.INT)
@click.option('-gpercent', '--gappercent', type=click.INT)
@click.option('-stepdown', type=click.INT)
@click.option('-probe', type=click.INT)
@click.option('-start', type=click.INT)
@click.option('-end', type=click.INT)
@click.option('-step', type=click.INT)
@click.option('-greedy', is_flag=True)
@click.option('-excl', is_flag=True)
@click.option('-noquerylog', is_flag=True)

def output(strand:str, length:int, mismatch:int, cutoff:int,
           threads:int, greedy:bool,excl:bool,noquerylog:bool, 
           gap:int=500,
           gappercent:int|None = None, 
           stepdown:int|None = None, 
           probe:int|None=None,
           start:int|None=None, 
           end:int|None =None, 
           step:int|None=None,
           currentfolder = './data/', # can be adapted so the code can be run in other folders
           cutoff_cost:float = 1e6,
           cutoff_oligo:float = 10, # max allowed cost for a single oligo
           finished:bool = False,
           count:int = 1,
           completelyfailed:list = [],
           )->None:
    # initialization
    
    logdir = os.path.join(currentfolder,"logfiles/")
    try:
        os.mkdir(logdir)
    except FileExistsError:
        pass

    now = datetime.now()
    nowstring = now.strftime("%Y_%m_%d_%H:%M:%S")
    logpath = os.path.join(logdir,f'cycling_query_{nowstring}.log')

    logging.basicConfig(filename=logpath, level=logging.DEBUG)
    logging.info(f"Starting probe query: "+nowstring)
    logging.info(f"User-set parameters:")
    logging.info(f"FISH type                : {strand}")
    logging.info(f"Oligo length             : {length}")
    logging.info(f"HUSH mismatches          : {mismatch}")
    logging.info(f"Max nb of off-targets    : {cutoff}")
    if(excl):
        logging.info(f"Masking probe region from HUSH runs.")
    

    outprobes = os.path.join(currentfolder,"final_probes/")
    try:
        os.mkdir(outprobes)
    except FileExistsError:
        pass
    # TO DO: check already completed probes and skip

    # parameters
          # max total cost of a probe. Exclude probe if even one oligo has a prohibitive cost
    if (not gappercent):
        cutoff_d_pc = 10
    else:
        cutoff_d_pc = gappercent    
    #cutoff_d_pc = 10        # max distance between 2 consecutive oligos, as a % of the total probe length
    cutoff_d = gap          # max distance between 2 consecutive oligos, in nucleotides 
    #cutoff_d = 500          # max distance between 2 consecutive oligos, in nucleotides 
    
    if (not stepdown):
        stepdown = 1                # number of oligos to decrease size of probe with, if no valid probe could be found with the current size       

    logging.info(f"Encoded parameters:")
    logging.info(f"Max probe cost                               : {cutoff_cost}")
    logging.info(f"Max dist. between oligos (% probe length)    : {cutoff_d_pc}")
    logging.info(f"Max dist. between oligos (nt)                : {cutoff_d}")
    logging.info(f"Max single oligo cost                        : {cutoff_oligo}")

    # start process with full ROI list
    roilist = os.path.join(currentfolder,'rois/all_regions.tsv')
    rdroi = pd.read_csv(roilist,sep="\t",header=0)
    
    # moved into function args as they are independently initialized 
    #finished = False
    #count = 1
    #completelyfailed = []

    if ((not start) or (not end) or (not step)):          # check if the user has provided all three parameters to sweep different oligo numbers
        sweep = False  
    else:
        sweep = True
        oligorange = range(start,end,step)
    if (not probe):
        toprocess = rdroi
    else:
        toprocess = rdroi[rdroi.window_id == probe]      
          
    while (not finished):

        logging.warning(f"Probe generation round {count}.")

        print(f"Probe generation round {count}.")

        # EMPTY SELECTED_PROBES FOLDER + MAKE SURE PROBE QUERY FOLDERS DON'T COLLIDE
        try:
            shutil.rmtree(os.path.join(currentfolder , "probe_candidates"))
        except:
            pass
        try:
            shutil.rmtree(os.path.join(currentfolder , "selected_probes")) 
        except:
            pass       

        os.mkdir(os.path.join(currentfolder,"probe_candidates/"))
        os.mkdir(os.path.join(currentfolder,"selected_probes/"))
        
        toprocessRoi = toprocess.window_id.to_list()
        toprocessOligos = toprocess.window.to_list()

        # update the probe databases with removed oligos
        #filterdatabase(currentfolder,cutoff_oligo,toprocessRoi)    

        filterdatabase_par(currentfolder,cutoff_oligo,threads,toprocessRoi) 

        if excl:
            flag = "-e"
        else:
            flag=""    

        for n in tqdm(range(len(toprocess)),"Generating probe candidates..."):
            # retrieve ROI number from ROI name
            roinumber = toprocessRoi[n]
            oligos = toprocessOligos[n]
            timestamp = datetime.now()
            ts_string = timestamp.strftime("%Y%m%d_%H%M%S")

            if oligos <= 0:         # no probe was found at any length
                completelyfailed.append(roinumber)
                logging.warning(f"No probe could be found for ROI "+str(roinumber)+". Proceeding with the other probes.")
                continue

            # if the user provided start/end/step, use as range of oligo numbers to design probes for the first time!
            if sweep:
                querylogpath = logdir + "query_roi_"+str(roinumber)+"_oli_sweep_round_"+str(count)+"_" + ts_string + ".txt"
                timeout=9999999
                with tqdm_joblib(tqdm(desc="Sweeping oligo counts in region "+str(roinumber), total=len(oligorange))) as progress_bar:
                    Parallel(n_jobs=threads,timeout=timeout)(delayed(probequery)(length,strand,roinumber,oligos,querylogpath,greedy,noquerylog) for oligos in oligorange)
            else:
                querylogpath = logdir + "query_roi_"+str(roinumber)+"_oli_"+str(oligos)+"_round_"+str(count)+"_" + ts_string + ".txt"
                # use as input for probe query (only process remaining ROIs)
                probequery(length,strand,roinumber,oligos,querylogpath,greedy,noquerylog)

        # select best probes
        print(f"Selecting probes...")
        if(sweep):
            selection = selectprobes(currentfolder, toprocessRoi, [start]*len(toprocessRoi), cutoff_cost, cutoff_d, cutoff_d_pc) 
        else:
            selection = selectprobes(currentfolder, toprocessRoi, toprocessOligos, cutoff_cost, cutoff_d, cutoff_d_pc) 
    
        failedlist = selection[selection.success == 0].index.to_list()
        logging.info(f'Length of failedlist: '+str(len(failedlist)))
        
        selectedlist = selection[selection.success == 1].index.to_list()

        if(len(selectedlist)>0):
            # check off-target homology with HUSH
            timestamp = datetime.now()
            ts_string = timestamp.strftime("%Y%m%d_%H%M%S")
            hushlogpath = logdir + "hush_roi_round_"+str(count)+"_" + ts_string + ".txt"
            print(f"Checking the oligos with (old)HUSH...")
            with open(hushlogpath,'w') as f:
#            subprocess.run("./validation_oldHUSH_BLAST.sh -L "+str(length)+" -m "+str(mismatch)+" -t "+str(threads)+flag+" > "+hushlogpath, shell=True,check=True)
                subprocess.run(["prb","validation_oldHUSH_BLAST", '-L', str(length),"-m",str(mismatch),"-t",str(threads),flag],stdout=f)
            print(f"Removing poor oligos from database")

        # apply results from HUSH to exclude poor oligos
        rerunlist = feedback(currentfolder,outprobes,count,cutoff,logpath)

        combinedlist = np.unique(failedlist+rerunlist)

        if (len(combinedlist) == 0):
            finished = True
            logging.info(f"Done! :)")
            if(len(completelyfailed)>0):
                logging.info(f"No probe could be found for the following regions: "+''.join(str(e)+", " for e in completelyfailed)+".")
            break
        else:
            print(f""+str(len(combinedlist))+" probes need to be re-run.")
            
            # set the correct oligo numbers for the next iteration
            #  
            # if a probe was selected but rejected by HUSH, re-use the selected number of oligos
            maskrerun = [(toprocess.loc[k,'window_id'] in rerunlist) for k in toprocess.index.to_list()] 
            rerunrois = toprocess[maskrerun]
            if (len(rerunrois)>0):
                toprocess.loc[maskrerun,"window"] = [int(selection.loc[k,'oligos']) for k in rerunrois.window_id.to_list()]
                #toprocess["window"][maskrerun] = [int(selection.loc[k,'oligos']) for k in rerunrois.window_id.to_list()]
            # for probes for which no valid probe could be constructed, reduce number of oligos for next iteration
            maskfailed = [(toprocess.loc[k,'window_id'] in failedlist) for k in toprocess.index.to_list()] 
            failedrois = toprocess[maskfailed]
            if (len(failedrois)>0):
                toprocess.loc[maskfailed,"window"] = [int(selection.loc[k,'oligos']-stepdown) for k in failedrois.window_id.to_list()]
                #toprocess["window"][maskfailed] = [int(selection.loc[k,'oligos']-stepdown) for k in failedrois.window_id.to_list()]
            # only keep rois that need to be re-run
            maskcombined = [(toprocess.loc[k,'window_id'] in combinedlist) for k in toprocess.index.to_list()] 
            toprocess = toprocess[maskcombined]

            count = count+1  
            sweep = False   


# -----------------------------------------------------------------------------------------------------------------------      
# -----------------------------------------------------------------------------------------------------------------------            


def selectprobes(input_folder, toprocessroi, toprocessoligos, cutoff_cost, cutoff_d, cutoff_d_pc):

    # retrieve probe queries
    selectedfolder = input_folder + 'selected_probes/'
    try:
        os.mkdir(selectedfolder)
    except FileExistsError:
        pass
    # identify probe files
    pattern = input_folder+"probe_candidates/**/probe_*.tsv"   #probelet
    #pattern = "data/**/probe*/**oligos.tsv"    #ifpd2 query
    #pattern = "data/**/*.best_probe.tsv"   #jupyternb
    filenames = glob.glob(pattern,recursive=True)

    roilist = input_folder+'rois/all_regions.tsv'
    rdroi = pd.read_csv(roilist,sep="\t",header=0)

    probelist = pd.DataFrame(columns=['fullpath', 'folder', 'probe_set', 'roi', 'chr', 'probe_start', 'probe_end', \
            'nOligos', 'pw', 'probe_size', 'region_size', 'coverage', \
            'centrality', 'd_max_pcregion', 'd_max_pcprobe',\
            'd_mean', 'd_min', 'd_max', 'd_std', \
            'tm_range', 'tm_mean', 'tm_std', \
            'gc_range', 'gc_mean', 'gc_std', \
            'mean_closestMM', 'min_closestMM', 'max_closestMM', 'std_closestMM', \
            'mean_cumulMM', 'min_cumulMM', 'max_cumulMM', 'std_cumulMM', \
            'mean_oligo_cost', 'min_oligo_cost', 'max_oligo_cost', 'std_oligo_cost' ,\
            'sum_inv_cost'])
            #'mean_sumMM', 'min_sumMM', 'max_sumMM', 'std_sumMM'   ])
        

    # read probe files
    if len(filenames)>0:
        for file in tqdm(filenames, "Summarizing probe candidates..."):
            rd = pd.read_csv(file,sep="\t")
            last = len(probelist)+1
            
            # find coordinates of the corresponding roi
            current_roi = rd.name[0][4:]
            for k in range(len(rdroi)):
                if str(rdroi.window_id[k]) == current_roi:
                    current = k

            roi_start = rdroi.Window_start[current]
            roi_end = rdroi.Window_end[current]
            roi_center = ((roi_start+roi_end)/2)-roi_start  #adjusted roi center

        # calculate summary measures and add as last row

            inv_cost = np.divide(np.ones(len(rd.name.tolist())),rd.oligo_cost)

        # replace off_target_no by min_mismatch for new queries
            probe_start = min(rd.start[:])
            probe_end = max(rd.end[:])
            probe_center = ((probe_start+probe_end)/2)-roi_start    #adjusted probe center

            filesplit = file.split("/")         # split numbers will have to be adjusted for running in different folders

            probelist.loc[last] = [file, filesplit[3], filesplit[4], int(rd.name[0][4:]), rd.chromosome[0], probe_start, probe_end, \
                len(rd), float(filesplit[4][-8:-4]), probe_end-probe_start+1, int(roi_end)-int(roi_start)+1, (probe_end-probe_start+1)/(int(roi_end)-int(roi_start)+1), \
                min(probe_center/roi_center,2-(probe_center/roi_center)), 100*(rd.start-rd.end.shift()).max()/(int(roi_end)-int(roi_start)+1), 100*(rd.start-rd.end.shift()).max()/(int(probe_end)-int(probe_start)+1),\
                (rd.start-rd.end.shift()).mean(), (rd.start-rd.end.shift()).min(), (rd.start-rd.end.shift()).max(), (rd.start-rd.end.shift()).std(), \
                max(rd.Tm)-min(rd.Tm), np.mean(rd.Tm), np.std(rd.Tm), \
                max(rd.gc_content)-min(rd.gc_content), np.mean(rd.gc_content), np.std(rd.gc_content), \
                np.mean(rd.off_target_no), min(rd.off_target_no), max(rd.off_target_no), np.std(rd.off_target_no), \
                np.mean(rd.off_target_sum), min(rd.off_target_sum), max(rd.off_target_sum), np.std(rd.off_target_sum), \
                np.mean(rd.oligo_cost), min(rd.oligo_cost), max(rd.oligo_cost), np.std(rd.oligo_cost), \
                inv_cost.sum()]
            # stat.mean(rd.off_target_sum), min(rd.off_target_sum), max(rd.off_target_sum), stat.stdev(rd.off_target_sum)]  
        
        probelist.loc[:,'fullpath'] = probelist.loc[:,'fullpath'].astype(str)

    selection = pd.DataFrame(columns=['roi','oligos','pw','success'])   # summary of selection results
    selection.set_index('roi',inplace=True)

    # one selected probe per ROI
    for n in tqdm(range(len(toprocessroi)),"Selecting optimal probes..."):
        roi = toprocessroi[n]
        minoligo = toprocessoligos[n]
        try:
            roiprobelist = probelist[(probelist.max_oligo_cost < cutoff_cost) & (probelist.d_max_pcprobe < cutoff_d_pc) & (probelist.d_max < cutoff_d) & (probelist.roi == roi)]
            filteredroilist = roiprobelist[roiprobelist.nOligos == roiprobelist.nOligos.max()]
            selectedprobeindex = filteredroilist[['pw']].idxmin()
            shutil.copy2(filteredroilist.loc[selectedprobeindex].fullpath.tolist()[0],selectedfolder)
            selection.loc[roi] = [roiprobelist.nOligos.max(),roiprobelist.loc[selectedprobeindex,'pw'].to_list()[0],1]
        except:
            logging.info(f"No valid probe was found for ROI #"+str(roi)+".")
            if (len(probelist[probelist.roi == roi]) > 0):
                selection.loc[roi] = [probelist[probelist.roi == roi].nOligos.min(),0,0]    # smallest successful number of oligos
            else:
                selection.loc[roi] = [minoligo,0,0]   # min oligo count that was unsuccessfully tested   

    return selection


# -----------------------------------------------------------------------------------------------------------------------      
# -----------------------------------------------------------------------------------------------------------------------            


def feedback(currentfolder,outfolder,count,cutoff,logpath):
    selectedfolder = currentfolder + 'selected_probes/'
    # identify probe files
    hushpattern = currentfolder+"selected_probes/query_*.out"   # HUSH validation output
    hushnames = glob.glob(hushpattern,recursive=True)

    logging.info(f'Length of hushnames: '+str(len(hushnames)))
    #logging.info(f'Hushnames: '+' '.join(str(l) for l in hushnames))  

    dbpattern = currentfolder+"db_tsv/db.*.tsv"   # ROI oligo databases
    dbnames = glob.glob(dbpattern,recursive=True)

    rerunlist = []

    # for each result file from HUSH
    # read all rows and fetch HUSH score
    if(len(hushnames)>0):
        for hushfile in hushnames:

            #logging.info(f'Current hush file: '+hushfile)

            # identify DB file
            filesplit = hushfile.split('/')
            basename = filesplit[len(filesplit)-1]
            roiname = basename[12:basename.find('.')]      # extracted roi name
            dbtest = [db.find("."+roiname+".")>0 for db in dbnames]
            roiDb = list(compress(dbnames,dbtest))[0]      # identify the corresponding database file

            logging.info(f'Currently processing '+roiname+', round #'+str(count)+".")

            hush_fasta = open(hushfile, 'r')  
            hushscores = []
            lines = hush_fasta.readlines()
            for line in lines:
                linesplit = line.split(sep=", ")
                if len(linesplit) > 1:
                    hushscores.append(int(linesplit[1][:-1]))    

            filename = hushfile.replace('query_','')
            tsvfile = filename[:filename.find('.fa')]+".tsv"

            # retrieve oligos with poor HUSH score
            candidate = pd.read_csv(tsvfile,sep="\t",header=0)
            candidate = candidate.set_index('start')
            candidate['HUSH_score'] = hushscores
            exclude = candidate[candidate.HUSH_score > cutoff]

            if(len(exclude)>0):
                logging.info(f'Excluding '+str(len(exclude))+' oligos from the database.')
                # attribute prohibitive escafish score in the ROI oligo database
                roioligos = pd.read_csv(roiDb,sep="\t",header=0)
                cols = list(roioligos.columns.values)       # save the order of the columns to restore it later
                roioligos = roioligos.set_index('start')
                roioligos.loc[exclude.index,'oligo_cost'] = 1e10

                # export the updated database in place
                roioligos = roioligos.reset_index()
                roioligos = roioligos[cols]
                roioligos.to_csv(roiDb,index=False,sep="\t")

                rerunlist.append(int(roiname[4:]))  # the probe will have to be queried again from the updated oligo database

            else:
                logging.info(f'No oligos were excluded.')  
                # move probe to final selection folder
                tsvname = basename[6:basename.find('.fa')]+".tsv"
                shutil.move(selectedfolder + tsvname,outfolder)
                shutil.move(selectedfolder + basename,outfolder)    # also keep the .out file (all other files will be deleted)      
                    

    logging.info(f'Length of rerunlist: '+str(len(rerunlist)))
    logging.info(f'Rerunlist: '+' '.join(str(l) for l in rerunlist))

    return rerunlist


# -----------------------------------------------------------------------------------------------------------------------      
# -----------------------------------------------------------------------------------------------------------------------            


def filterdatabase(currentfolder,cutoff,toprocessRoi):
    # generate a filtered copy of the oligo database for each ROI
    # filter by removing oligos over a certain threshold cost

    dbfolder = currentfolder + 'db_tsv/'
    # identify database files
    dbpattern = dbfolder+"db.*.tsv"   # HUSH validation output
    dbnames = glob.glob(dbpattern,recursive=True)

    dbparse = [re.split('roi_',db,maxsplit=1)[1] for db in dbnames]
    dbrois = [int(re.split('\.',db,maxsplit=1)[0]) for db in dbparse]

    filtereddblist = [dbnames[k] for k in range(len(dbrois)) if dbrois[k] in toprocessRoi]

    # for each database file
    # open the updated (full) database and export a filtered version without discarded oligos       

    for db in tqdm(filtereddblist,"Preparing filtered oligo databases..."):
        # retrieve oligos with poor HUSH score
        oligodb = pd.read_csv(db,sep="\t",header=0)
        filtereddb = oligodb[oligodb.oligo_cost < cutoff]
        filtereddb.to_csv(db+".filt",index=False,sep="\t")

def filterdatabase_par(currentfolder,cutoff,threads,toprocessRoi):
    # generate a filtered copy of the oligo database for each ROI
    # filter by removing oligos over a certain threshold cost

    dbfolder = currentfolder + 'db_tsv/'
    # identify database files
    dbpattern = dbfolder+"db.*.tsv"   # HUSH validation output
    dbnames = glob.glob(dbpattern,recursive=True)

    dbparse = [re.split('roi_',db,maxsplit=1)[1] for db in dbnames]
    dbrois = [int(re.split('\.',db,maxsplit=1)[0]) for db in dbparse]

    filtereddblist = [dbnames[k] for k in range(len(dbrois)) if dbrois[k] in toprocessRoi]

    # for each database file
    # open the updated (full) database and export a filtered version without discarded oligos
    with tqdm_joblib(tqdm(desc="Filtering oligo databases", total=len(filtereddblist))) as progress_bar:
        Parallel(n_jobs=threads)(delayed(filterdb)(db,cutoff) for db in filtereddblist)
 

def filterdb(db,cutoff):
     # retrieve oligos with poor HUSH score
    oligodb = pd.read_csv(db,sep="\t",header=0)
    filtereddb = oligodb[oligodb.oligo_cost < cutoff]
    filtereddb.to_csv(db+".filt",index=False,sep="\t")       

# -----------------------------------------------------------------------------------------------------------------------      
# -----------------------------------------------------------------------------------------------------------------------            

def probequery(length,strand,roi,oligos,logpath,greedy,noquerylog):
    suffix = ""
    if(greedy): 
        suffix = "-g"
    if noquerylog:    
        #subprocess.run("./probe-query.sh -s "+strand+" -e "+str(roi)+" -o "+str(oligos)+suffix+"> /dev/null 2>&1", shell=True)
        subprocess.run(["prb","probe-query","-s",strand,"-e",str(roi),"-o",str(oligos),suffix], stdout=None)
    else:
        with open(logpath,'w') as f:
        #subprocess.run("./probe-query.sh -s "+strand+" -e "+str(roi)+" -o "+str(oligos)+suffix+" > "+logpath+" 2>&1", shell=True)
            subprocess.run(["prb","probe-query","-s",strand,"-e",str(roi),"-o",str(oligos),suffix],stderr=subprocess.STDOUT,stdout=f)

# -----------------------------------------------------------------------------------------------------------------------      
# -----------------------------------------------------------------------------------------------------------------------            
def cycling_query():
    output()
    logging.shutdown()
    return

if __name__ == '__main__':
    cycling_query()



from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
import argparse
import os

'''
Split a CHM13 T2T (telomere-to-telomere) genome fasta file into chromosomes.
The chromosome id is extracted from the fasta title.
The output files are named as "save_loc/prefix.chromosome.chr_id.fa"
Also, FASTA sequences are transformed to uppercase.
'''

argparser = argparse.ArgumentParser(description='Split a CHM13 T2T (telomere-to-telomere) genome fasta file into chromosomes.')
argparser.add_argument('-g','--genome', help='Input genome fasta file.',default='GCA_009914755.4_T2T-CHM13v2.0_genomic.fna')
argparser.add_argument('-p','--prefix', help='Prefix for the output files.', default='CHM13.T2T')
argparser.add_argument('-s','--save_loc', help='Location to save the output chromosome files.', default='.')
args = argparser.parse_args()

def split_fasta(genome_fasta:str, prefix:str='chm13',save_loc:str='.')->None:
    '''
    Read a genome fasta file and split it into chromosomes.
    The chromosome id is extracted from the fasta title.
    The output files are named as "save_loc.prefix.chromosome.chr_id.fa"
    '''
    with open(genome_fasta, 'r') as f:
        for record in SeqIO.parse(f, 'fasta'):
            title = record.description
            sequence = record.seq.upper() # Convert to uppercase
            words = title.split(' ')
            if not 'mitochondrion' in title: # Not a mitochondrial chromosome
                chr_id = words[-1] # last word is the chromosome id
            else : # Mitochondrial chromosome
                chr_id = 'M' 

            fname = os.path.join(save_loc, f'{prefix}.chromosome.{chr_id}.fa') # Output file name
            with open(fname, 'w') as out:
                # Create a new SeqRecord for the sequence
                print(f'\033[92mWriting chromosome {chr_id} to {fname}\033[0m',end=' ')
                new_record = SeqRecord(sequence, id=record.id, description=title)
                # Write the SeqRecord to the file
                SeqIO.write(new_record, out, 'fasta')   # 60 characters per line
                print('\u2705')


if __name__ == '__main__':
    split_fasta(args.genome, args.prefix, args.save_loc)
    test_script = 'python3 split_it.py -g GCA_009914755.4_T2T-CHM13v2.0_genomic.fna -p chm13 -s ./'
    test_for_hpc = 'python3 split_it.py -g data/ref/GCA_009914755.4_T2T-CHM13v2.0_genomic.fna -p CHM13.T2T -s data/ref/'



# Python code to download the chromosomes and genome of a species from Ensembl FTP server
import requests
import os
import gzip
import argparse
from bs4 import BeautifulSoup
from split_GRC import split_fasta
from tqdm import tqdm

# Human (homo_sapiens) and Mouse (mus_musculus) genomes respectively
GRCh = "https://ftp.ensembl.org/pub/release-{release}/fasta/homo_sapiens/dna/Homo_sapiens.GRCh{build}.dna.chromosome.{chr}.fa.gz"
GRCm = "https://ftp.ensembl.org/pub/release-{release}/fasta/mus_musculus/dna/Mus_musculus.GRCm{build}.dna.chromosome.{chr}.fa.gz"

# available only mouse and human
available = ['homo_sapiens','mus_musculus']
human = ['homo_sapiens']
mouse = ['mus_musculus']

human_all_chr: list[int|str] = list(range(1,23))+['X','Y','MT'] # list of human chromosomes
mouse_all_chr: list[int|str] = list(range(1,19))+['X','Y','MT'] # list of mouse chromosomes

#args
argparser = argparse.ArgumentParser(prog='download ensemble genome',description="download ensemble genome")
argparser.add_argument('-s','--species',type=str,default='homo_sapiens',choices=["homo_sapiens","mus_musculus"])
argparser.add_argument('-b','--build',type=int,default=38,help="the build number of the genome")
argparser.add_argument('-r','--release',default=109,help="release number of the build")
argparser.add_argument('-d','--dir',default='data/ref',help='destination directory')
argparser.add_argument('-f','--filename',default=None,help="give a specific name to the downloaded file ")
argparser.add_argument('-k','--keep',action='store_true',help="whether to keep gzip files")
argparser.add_argument('-split',action='store_true',help='whether to split into chromosomes')
args = argparser.parse_args()


def get_ensembl_release(species:str='Homo_sapiens')->int:
    '''
    get the latest release number for ensemble
    '''

    # get the website
    ws = requests.get(f"https://www.ensembl.org/{species}/Info/Index")

    # get the tile
    soup = BeautifulSoup(ws.content, 'html.parser')
    title = soup.find('title').string

    # get the species and build number
    species = title.split()[0]
    latest_release = title.split()[-1]

    print(f"\033[1m\033[94m build (latest) {latest_release} for {args.species} \033[0m")

    return latest_release

def download_chr(chr_folder:os.PathLike="data/ref",
                 chr:int|str=17,
                 release:int|str=109,
                 build:int=38,
                 species:str="homo_sapiens",
                 unzip:bool=True,
                 remove_gz:bool=False,
                 file_name:str|None=None,
                 url_chr:str|None=None)->None:
    
    """
    Downloads the specified chromosome of a genome from Ensembl FTP server. Requires internet connection...

    Args:
        chr_folder (os.PathLike, optional): 
            Path to the folder where the chromosome file will be saved. Defaults to "data/ref".
        chr (int|str, optional): 
            Chromosome number or identifier. Defaults to 17.
        release (int|str, optional): 
            Release version of the genome. Defaults to 109.
        build (int, optional):
            Build version of the genome. Defaults to 38.
        species (str, optional):
            Species of the genome (homo_sapiens or mus_musculus). Defaults to "homo_sapiens".
        unzip (bool, optional): 
            Flag to indicate whether to unzip the downloaded file. Defaults to True.
        remove_gz (bool, optional): 
            Flag to indicate whether to remove the downloaded gzipped file. Defaults to False.

    Returns:
        None
    """

    # matching species with the link
    species = species.lower()
    if url_chr is None:
        if species in human:
            url_chr = GRCh
        elif species in mouse:
            url_chr = GRCm
        else:
            print(f"species should be one of {available}")
            print(f"You entered {species}")
            return
        print(f"\033[91m downloading from\033[0m\n{url_chr} ...")
        url_chr = url_chr.replace('{release}',f'{release}'
                                    ).replace('{chr}',f'{chr}'
                                        ).replace('{build}',f'{build}') # placing release and chr numbers
    else:
        print(f"downloading from {url_chr} ...")
        if 'assembly' in url_chr:
            print(f"\033[91mREFERENCE genome, this may take some time: 5-15 mins depending on your internet speed\033[0m")
        pass # url_chr = url_chr

    if file_name is None:
        file_name = url_chr.split('/')[-1] # last part of the link will be the file name
    else:
        pass # file_name = file_name
    
    # creates chr_folder if not exists
    if not os.path.exists(chr_folder):
        try: # just in case the directories do not exist
            os.makedirs(chr_folder,exist_ok=True)
        except:
            print(f'parent directory {os.path.dirname(chr_folder)} does not exist')
            print('exiting...')
            return
    
    chr_path = os.path.join(chr_folder,file_name)
    
    # read the link to a variable chr_fasta_gz
    if not os.path.exists(chr_path): # do not overwrite if it exists
        
        #try:
        if True:
            chr_fasta_gz = requests.get(url_chr,stream=True)
            total_size = int(chr_fasta_gz.headers.get('content-length', 0))
            block_size = 1024 #1 Kbyte
            with tqdm(total=total_size, unit='B', unit_scale=True) as progress_bar:
                with open(chr_path,"wb") as f:
                    for data in chr_fasta_gz.iter_content(block_size):
                        progress_bar.update(len(data))
                        f.write(data)
            #open(chr_path,"wb").write(chr_fasta_gz.content)# write chr_fasta_gz into the chromosome path
        #except:
        #    print("Link does not work:",url_chr)
        #    print("Exiting...")
         #   return
    else:
        print(f'file {chr_path} already exists. Skipping...')


    if unzip: # unzip the .gz file if requested
        chr_fasta = chr_path[:-3]  # Remove the '.gz' extension
        if not os.path.exists(chr_fasta): # do not overwrite chromosome fasta file if exists
            print(f'unzipping...')
            # Unzip .gz file and save the unzipped content to a new file
            open(chr_fasta, 'wb').write(gzip.open(chr_path, 'rb').read())
            print(f'unzipped file saved to {chr_fasta}')
        else:
            print(f'file {chr_fasta} already exists. Skipping...')

    if remove_gz: # remove the .gz file if requested
        if os.path.exists(chr_path):
            print(f'Removing gzipped file {chr_path} ...', end=' ')
            os.remove(chr_path)
            print('\u2705')
        else: # could already be removed
            print(f'file {chr_path} does not exist. Skipping...')


    return

def download_chr_list(list_chr:str|list = 'all',   
                      chr_folder:os.PathLike="data/ref",
                      release:int|str=109,
                      build:int=38,
                      species:str="homo_sapiens",
                      unzip:bool=True,
                      remove_gz:bool=False)->None:
    
    """
    Downloads the specified chromosome lists of a genome from Ensembl FTP server. 
        Requires internet connection...

    Args:
        list_chr (list[int|str], optional): 
            List of chromosome numbers or identifiers. Defaults to list(range(1,23))+['X','Y'].
        chr_folder (os.PathLike, optional): 
            Path to the folder where the chromosome file will be saved. Defaults to "data/ref".
        release (int|str, optional): 
            Release version of the genome. Defaults to 109.
        build (int, optional):
            Build version of the genome. Defaults to 38.
        species (str, optional):
            Species of the genome (homo_sapiens or mus_musculus). Defaults to "homo_sapiens".
        unzip (bool, optional): 
            Flag to indicate whether to unzip the downloaded file. Defaults to True.
        remove_gz (bool, optional): 
            Flag to indicate whether to remove the downloaded gzipped file. Defaults to False.

    Returns:
        None
    """
    if list_chr=="all":
        if species in human :
            list_chr = human_all_chr
        elif species in mouse:
            list_chr = mouse_all_chr
        else:
            print(f"species should be one of {available}")
            print(f"You entered {species}")
            return
    elif type(list_chr) == list or type(list_chr) == tuple:
        pass # 
    else:
        print('list_chr should either be "all", a list or tuple')
        print(f'You entered {type(list_chr)} type which is not comptatible..')
        return

    for chr in list_chr:
        print(f'downloanding... {species=} {chr=} {release=} {build=} ')
        download_chr(chr_folder=chr_folder,
                     chr=chr,
                     release=release,
                     species=species,
                     unzip=unzip,
                     remove_gz=remove_gz,
                     )
    return

def download_ref_genome(build:int=38,
                        release:int=109,
                        species:str="homo_sapiens",
                        chr_folder:os.PathLike="data/ref",
                        file_name:str='genome.fa.gz',
                        unzip:bool=True,
                        remove_gz:bool=False)->None:
    
    if species in human:
        url_chr = GRCh.replace('{release}',f'{release}').replace('{build}',f'{build}')
        
    elif species in mouse:
        url_chr = GRCm.replace('{release}',f'{release}').replace('{build}',f'{build}')


    url_chr = url_chr.replace('chromosome.{chr}','primary_assembly')
    download_chr(chr_folder=chr_folder,
                 url_chr=url_chr,
                 unzip=unzip,
                 file_name=file_name,
                 remove_gz=remove_gz)
    return

def main():

    if args.release != 'latest':
        download_ref_genome(
            build=args.build,
            release=args.release,
            species=args.species,
            chr_folder=args.dir,
            file_name=args.filename,
            unzip=True,
            remove_gz=not args.keep
        )
    else:
        release = get_ensembl_release(species=args.species.capitalize())
        download_ref_genome(
            build=args.build,
            release=release,
            species=args.species,
            chr_folder=args.dir,
            file_name=args.filename,
            unzip=True,
            remove_gz=not args.keep
        )

    if args.split:
        import glob
        fasta = glob.glob('data/ref/*.dna.primary_assembly.fa')[0]
        split_fasta(genome_fasta=fasta,prefix=None,save_loc='.') # split the genome into chromosomes
        os.rename(fasta,'genome.fa')

if __name__ == "__main__":
    main()




#!/usr/bin/python3
#set -e

# Generate default masked-out regions, corresponding to the region coordinates as provided by the user.
# Used to avoid excluding region-specific repetitive sequences when running (n)HUSH.

# No arguments required.

from cmath import exp
import numpy as np
from tabulate import tabulate as tab
import os
from tqdm import tqdm
import pandas as pd 
import sys
import contextlib
from tqdm import tqdm


def generate_exclude(currentfolder:os.PathLike = './data')->None:

    # paths
    roilist = currentfolder+'/rois/all_regions.tsv'
    out = currentfolder+'/exclude/'

    rd = pd.read_csv(roilist,sep="\t",header=0)
    ROIcount = len(rd)

    try:
        os.mkdir(out)
    except FileExistsError:
        print("The data/exclude folder already exists, risk of overwriting manual input! Exiting.")
    else:    
        
        # for all ROI in the region list, create an exclusion file with the region coordinates
        for roi in tqdm(range(ROIcount),'Generating exclusion maps...'):
            fixedchr = rd.ref[roi]+".chromosome."+rd.chrom[roi][3:]
            exporttable = pd.DataFrame(columns=['chrom','chromStart','chromEnd'])
            exporttable.loc[roi]=[fixedchr,rd.Window_start[roi],rd.Window_end[roi]]
            filename = 'excl_roi_'+str(rd.window_id[roi])+'.bed'
            exporttable.to_csv(out+filename,index=False,sep="\t") 

    return

if __name__ == "__main__":
    generate_exclude()


#!/usr/bin/python3

from Bio.SeqIO.FastaIO import SimpleFastaParser
from ifpd2q.scripts.extract_kmers import main as extract
import pandas as pd
import os
import sys
from tqdm import tqdm
from tabulate import tabulate


def split_fasta(file:os.PathLike,outfolder:os.PathLike)->None:

    with open(file) as handle:
        for values in SimpleFastaParser(handle):
            if not ".fa" in file: # just in case they don't add the extension
                out = open(outfolder+values[0]+".fa", 'w')
            else:
                out = open(outfolder+values[0], 'w')
            out.write('>'+values[0]+'\n'+values[1])
            out.close()

    return

if __name__ == "__main__":
    file = sys.argv[1]
    outfolder = sys.argv[2]
    split_fasta(file=file,outfolder=outfolder)


#!/usr/bin/python3

from Bio.SeqIO.FastaIO import SimpleFastaParser
import numpy as np
import pandas as pd
import os
import glob
import sys
from tqdm import tqdm 
from tabulate import tabulate
import statistics as stat
import shutil
from itertools import compress

# for each result file from HUSH
# read all rows and fetch HUSH score
# read corresponding probe candidate tsv file
# attribute HUSH score to tsv file rows

def HUSH_feedback(currentfolder:os.PathLike = './data/',
                  cuttoff:int=99,
                  wipe:bool=True)->None:
    
    ## TO FIX:
    # cuttoff is not implemented
    # wipe is not implemented
    # selectedfolder is not implemented
    
    selectedfolder = currentfolder + 'selected_probes/'
    # identify probe files
    hushpattern = currentfolder+"selected_probes/query_*.out"   # HUSH validation output
    hushnames = glob.glob(hushpattern,recursive=True)

    dbpattern = currentfolder+"db_tsv/db.*.tsv"   # ROI oligo databases
    dbnames = glob.glob(dbpattern,recursive=True)

    # for each result file from HUSH
    # read all rows and fetch HUSH score
    for hushfile in hushnames:
        #print(f'Current: '+hushfile)
        # identify DB file
        filesplit = hushfile.split('/')
        basename = filesplit[len(filesplit)-1]
        basename = basename[12:basename.find('.')]      # extracted roi name
        dbtest = [db.find(basename)>0 for db in dbnames]
        roiDb = list(compress(dbnames,dbtest))[0]

        print(f'Currently processing '+basename+'.')

        hush_fasta = open(hushfile, 'r')  
        hushscores = []
        lines = hush_fasta.readlines()
        for line in lines:
            linesplit = line.split(sep=", ")
            if len(linesplit) > 1:
                hushscores.append(int(linesplit[1][:-1]))    

        filename = hushfile.replace('query_','')
        tsvfile = filename[:filename.find('.fa')]+".tsv"

        #print(f'TSV file: '+tsvfile)

        # retrieve oligos with poor HUSH score
        candidate = pd.read_csv(tsvfile,sep="\t",header=0)
        candidate = candidate.set_index('start')
        candidate['HUSH_score'] = hushscores
        exclude = candidate[candidate.HUSH_score > cutoff]

        finished = True

        if(len(exclude)>0):
            print(f'Excluding '+str(len(exclude))+' oligos from the database.')
            # attribute prohibitive escafish score in the ROI oligo database
            roioligos = pd.read_csv(roiDb,sep="\t",header=0)
            cols = list(roioligos.columns.values)       # save the order of the columns to restore it later
            roioligos = roioligos.set_index('start')
            roioligos.loc[exclude.index,'oligo_cost'] = 1e10

            # export the updated database in place
            roioligos = roioligos.reset_index()
            roioligos = roioligos[cols]
            roioligos.to_csv(roiDb,index=False,sep="\t")

            finished = False
        else:
            print(f'No oligos were excluded.')    

        if (not finished):
            print("Not completed. Run again!")

    return

if __name__ == "__main__":
    # for Command Line use
    if(len(sys.argv)>2):
        cutoff = sys.argv[1]
        wipe = sys.argv[2]
        HUSH_feedback(cuttoff=cutoff,wipe=wipe) # wipe default is True
    if(len(sys.argv)>1):
        cutoff = sys.argv[1]
        HUSH_feedback(cuttoff=cutoff)
    else:    
        cutoff = 99         # max HUSH score
        HUSH_feedback(cuttoff=cutoff) # cuttoff is already 99 by default in the function
        
        
        



from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
import argparse
import os

'''
Split a GRC genome fasta file into chromosomes.
The chromosome id is extracted from the fasta title.
The output files are named as "save_loc/prefix.chromosome.chr_id.fa"
Also, FASTA sequences are transformed to uppercase.
'''
all_chr: list[int|str] = list(range(1,23))+['X','Y','MT'] # list of human chromosomes
all_chr = [str(c) for c in all_chr]

def split_fasta(genome_fasta:str, prefix:str|None=None,save_loc:str='.')->None:
    '''
    Read a genome fasta file and split it into chromosomes.
    The chromosome id is extracted from the fasta title.
    The output files are named as "save_loc.prefix.chromosome.chr_id.fa"
    '''

    if prefix is None:
        position = genome_fasta.rfind('.primary_assembly')
        prefix = genome_fasta[:position]
  
    with open(genome_fasta, 'r') as f:
        for record in SeqIO.parse(f, 'fasta'):
            title = record.description
            sequence = record.seq.upper() # Convert to uppercase
            words = title.split(' ')

            chr_id = words[0] # first word is the chromosome id
            if chr_id not in all_chr:
                continue

            fname = os.path.join(save_loc, f'{prefix}.chromosome.{chr_id}.fa') # Output file name
            with open(fname, 'w') as out:
                # Create a new SeqRecord for the sequence
                print(f'\033[92mWriting chromosome {chr_id} to {fname}\033[0m',end=' ')
                new_record = SeqRecord(sequence, id=record.id, description=title)
                # Write the SeqRecord to the file
                SeqIO.write(new_record, out, 'fasta')   # 60 characters per line
                print('\u2705')
        
    return

if __name__ == '__main__':
    argparser = argparse.ArgumentParser(description='Split a GRC genome fasta file into chromosomes.')
    argparser.add_argument('-g','--genome', help='Input genome fasta file.',default='data/ref/Homo_sapiens.GRCh38.dna.primary_assembly.fa')
    argparser.add_argument('-p','--prefix', help='Prefix for the output files.', default=None)
    argparser.add_argument('-s','--save_loc', help='Location to save the output chromosome files.', default='.')
    args = argparser.parse_args()
    split_fasta(args.genome, args.prefix, args.save_loc)
